/*
 * ng_utp.c
 */

/*-
 * Copyright (c) 2012 Rozhuk Ivan <rozhuk.im@gmail.com>
 * All rights reserved.
 * 
 * Subject to the following obligations and disclaimer of warranty, use and
 * redistribution of this software, in source or object code forms, with or
 * without modifications are expressly permitted by Whistle Communications;
 * provided, however, that:
 * 1. Any and all reproductions of the source or object code must include the
 *    copyright notice above and the following disclaimer of warranties; and
 * 2. No rights are granted, in any manner or form, to use Whistle
 *    Communications, Inc. trademarks, including the mark "WHISTLE
 *    COMMUNICATIONS" on advertising, endorsements, or otherwise except as
 *    such appears in the above copyright notice or in the software.
 * 
 * THIS SOFTWARE IS BEING PROVIDED BY WHISTLE COMMUNICATIONS "AS IS", AND
 * TO THE MAXIMUM EXTENT PERMITTED BY LAW, WHISTLE COMMUNICATIONS MAKES NO
 * REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED, REGARDING THIS SOFTWARE,
 * INCLUDING WITHOUT LIMITATION, ANY AND ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT.
 * WHISTLE COMMUNICATIONS DOES NOT WARRANT, GUARANTEE, OR MAKE ANY
 * REPRESENTATIONS REGARDING THE USE OF, OR THE RESULTS OF THE USE OF THIS
 * SOFTWARE IN TERMS OF ITS CORRECTNESS, ACCURACY, RELIABILITY OR OTHERWISE.
 * IN NO EVENT SHALL WHISTLE COMMUNICATIONS BE LIABLE FOR ANY DAMAGES
 * RESULTING FROM OR ARISING OUT OF ANY USE OF THIS SOFTWARE, INCLUDING
 * WITHOUT LIMITATION, ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * PUNITIVE, OR CONSEQUENTIAL DAMAGES, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES, LOSS OF USE, DATA OR PROFITS, HOWEVER CAUSED AND UNDER ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF WHISTLE COMMUNICATIONS IS ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 *
 * Author: Rozhuk Ivan <rozhuk.im@gmail.com>
 *
 * $FreeBSD
 */


/*
 * started: 03.01.2012
 *
 * Netgraph "utp" node
 *
 * This node:
 * - Count: in/out packets, octets;
 * - Control: forward/drop
 * - Process: forward to default/specific node, for example ng_car
 * layer 2 packets
 * It can be used for count, control and shape IPv4+IPv6 users traffic
 */

#include <sys/param.h>
#include <sys/errno.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/mbuf.h>
#include <sys/queue.h>
#include <sys/lock.h>
#include <sys/rwlock.h>
#include <sys/mutex.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <vm/uma.h>

#include <net/ethernet.h>
#include <net/if.h>
#include <net/if_vlan_var.h>

#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_var.h>
#include <netinet/ip_options.h>
#include <netinet/ip6.h>
#include <netinet/if_ether.h>
#include <netinet/udp.h>

#include <machine/in_cksum.h>

#include <netgraph/ng_message.h>
#include <netgraph/netgraph.h>
#include <netgraph/ng_parse.h>

#include "ng_utp.h"
#include "uTP.h"

#ifdef NG_SEPARATE_MALLOC
MALLOC_DEFINE(M_NETGRAPH_UTP, "netgraph_utp", "netgraph utp node");
#else
#define M_NETGRAPH_UTP M_NETGRAPH
#endif




#define NET_PROTO_COUNT	2 /* IPv4, IPv6 */

//************************************************************************
//			Host L3-4 address
//************************************************************************
/*
 * Hash algorithm
 */
#define	HASH_SIZE	8 /* in bits */
#if HASH_SIZE == 8
#define HASH_BUCKETS	256
#define HASH2(addr)	(  ((const uint8_t *)(addr))[0]		\
			 ^ ((const uint8_t *)(addr))[1] )
#define HASH4(addr)	(  ((const uint8_t *)(addr))[0]		\
			 ^ ((const uint8_t *)(addr))[1]		\
			 ^ ((const uint8_t *)(addr))[2]		\
			 ^ ((const uint8_t *)(addr))[3] )
#define HASH16(addr)	(  ((const uint8_t *)(addr))[0]		\
			 ^ ((const uint8_t *)(addr))[1]		\
			 ^ ((const uint8_t *)(addr))[2]		\
			 ^ ((const uint8_t *)(addr))[3]		\
			 ^ ((const uint8_t *)(addr))[4]		\
			 ^ ((const uint8_t *)(addr))[5]		\
			 ^ ((const uint8_t *)(addr))[6]		\
			 ^ ((const uint8_t *)(addr))[7]		\
			 ^ ((const uint8_t *)(addr))[8]		\
			 ^ ((const uint8_t *)(addr))[9]		\
			 ^ ((const uint8_t *)(addr))[10]	\
			 ^ ((const uint8_t *)(addr))[11]	\
			 ^ ((const uint8_t *)(addr))[12]	\
			 ^ ((const uint8_t *)(addr))[13]	\
			 ^ ((const uint8_t *)(addr))[14]	\
			 ^ ((const uint8_t *)(addr))[15] )
#else /* 16 bits */
#define HASH_BUCKETS	65536
#define HASH2(addr)	(  ((const uint16_t *)(addr))[0] )
#define HASH4(addr)	(  ((const uint16_t *)(addr))[0]	\
			 ^ ((const uint16_t *)(addr))[1] )
#define HASH16(addr)	(  ((const uint16_t *)(addr))[0]	\
			 ^ ((const uint16_t *)(addr))[1]	\
			 ^ ((const uint16_t *)(addr))[2]	\
			 ^ ((const uint16_t *)(addr))[3]	\
			 ^ ((const uint16_t *)(addr))[4]	\
			 ^ ((const uint16_t *)(addr))[5]	\
			 ^ ((const uint16_t *)(addr))[6]	\
			 ^ ((const uint16_t *)(addr))[7] )
#endif
#define HASH_BUCKETS_COUNT	(HASH_BUCKETS * NET_PROTO_COUNT)

#define HASH_IPV4_HST(saddr, daddr, sport, dport) \
			(HASH4(saddr) | HASH4(daddr) | HASH2(sport) | HASH2(dport))
#define HASH_IPV6_HST(saddr, daddr, sport, dport) \
			(HASH16(saddr) | HASH16(daddr) | HASH2(sport) | HASH2(dport))


typedef struct utp_conn_s {
	TAILQ_ENTRY(utp_conn_s) next;
	uint32_t	connid;	   /* connection ID */
	uint8_t		hst_rev_state;/* L3-L4 conn state */
	uint8_t		pkt_ver;   /* uTP version */

	uint8_t		req_state; /* CONN_F_ST_DATA...CONN_F_ST_SYN bits */
	uint8_t		repl_state;/* CONN_F_ST_DATA...CONN_F_ST_SYN bits */
	//uint16_t	seq_nr;	   /* Sequence number */
	//uint16_t	ack_nr;	   /* Acknowledgment number */
	uint32_t	req_count; /* num of requests */
	uint32_t	repl_count;/* num of replyes */
	time_t		req_last;  /*  */
	time_t		repl_last; /*  */
} utp_conn_t, *utp_conn_p; /* 64 bytes */

#define	CONN_FLAG_VAL(val) (0x01 << val) /* Data packet. */
#define	CONN_F_ST_DATA	CONN_FLAG_VAL(ST_DATA) /* Data packet. */
#define	CONN_F_ST_FIN	CONN_FLAG_VAL(ST_FIN)  /* Finalize the connection. */
#define	CONN_F_ST_STATE	CONN_FLAG_VAL(ST_STATE) /* State packet. */
#define	CONN_F_ST_RESET	CONN_FLAG_VAL(ST_RESET) /* Terminate connection forcefully. */
#define	CONN_F_ST_SYN	CONN_FLAG_VAL(ST_SYN)  /* Connect SYN */
#define	CONN_F_CONNECTED (CONN_F_ST_DATA | CONN_F_ST_STATE | CONN_F_ST_SYN)

TAILQ_HEAD(utp_conn_head, utp_conn_s);



typedef struct hst_conn_s {
	TAILQ_ENTRY(hst_conn_s)	next;
	u_char		saddr[16]; /* L3 src IP v4/v6 addr */
	u_char		daddr[16]; /* L3 dst IP v4/v6 addr */
	uint16_t	sport; /* L4 src port */
	uint16_t	dport; /* L4 dst port */

	/* item data */
	volatile uint32_t lock_wait_count; /* num of threads wait for thos lock */
	struct mtx 	rw_lock; /* add/delete connection item lock */
	time_t		last_update;
	time_t		create_time;

	struct utp_conn_head conns_head; /* ports on host */
	uint32_t	conns_count;  /* ports count */
} hst_conn_t, *hst_conn_p; /* 128 bytes */

TAILQ_HEAD(hst_conn_head, hst_conn_s);


typedef struct hst_bucket_s {
	struct mtx 	rw_lock; /* add/delete host item lock */
	struct hst_conn_head hst_head;
} hst_bucket_t, *hst_bucket_p;


typedef struct hst_store_s {
	hst_bucket_t	buckets[HASH_BUCKETS_COUNT];
	uma_zone_t	hst_conn_zn; /* UMA allocator for hst_conn_t */
	uma_zone_t	utp_conn_zn; /* UMA allocator for utp_conn_t */
	struct callout	timer; /* one second periodic timer */
	volatile uint64_t hst_conns[NET_PROTO_COUNT]; /* Hosts IPv4/6 conns count */
	volatile uint64_t utp_conns[UTP_VERS_COUNT]; /* uTP v 0/1/... conns count */
} hst_store_t, *hst_store_p;


static int		hst_store_init(node_p node, hst_store_p h_store);
static void		hst_store_destroy(node_p node, hst_store_p h_store);
static uint32_t		hst_store_get_bucket_proto_idx(hst_store_p h_store,
			    hst_bucket_p bucket);
static hst_bucket_p	hst_store_proto_buckets(hst_store_p h_store, uint16_t l3_p);
static hst_bucket_p	hst_store_get_bucket(hst_store_p h_store, uint16_t l3_p,
			    u_char *saddr, u_char *daddr, uint16_t sport,
			    uint16_t dport, uint32_t *addr_len, uint32_t *proto_idx);
static hst_conn_p	hst_store_conn_alloc(hst_store_p h_store, uint16_t l3_p,
			    u_char *saddr, u_char *daddr,
			    uint16_t sport, uint16_t dport);
static void		hst_store_conn_free(hst_store_p h_store, hst_bucket_p bucket,
			    hst_conn_p hst_conn);
static utp_conn_p	hst_store_utp_conn_alloc(hst_store_p h_store,
			    hst_conn_p hst_conn, int hst_rev_state, int pkt_ver,
			    uint32_t connid);
static void		hst_store_utp_conn_free(hst_store_p h_store,
			    hst_conn_p hst_conn, utp_conn_p utp_conn);
static void		hst_store_timer(node_p node, hook_p hook, void *arg1,
			    int arg2);
static void		hst_store_flush_conn(hst_store_p h_store, uint16_t l3_p);
static hst_conn_p	hst_store_find_hst_conn(hst_store_p h_store, uint16_t l3_p,
			    u_char *saddr, u_char *daddr,
			    uint16_t sport, uint16_t dport, int *rev);
static utp_conn_p	hst_store_find_utp_conn(hst_conn_p hst_conn,
			    int hst_rev_state, int pkt_ver, uint32_t connid,
			    int *rev);
static int		hst_store_entryes4(hst_store_p h_store, u_char *buff,
			    uint32_t buff_size, uint32_t *entryes_cnt);



//************************************************************************
//			Node private data
//************************************************************************


typedef struct ng_utp_private {
	hst_store_t	h_store;
	ng_utp_nodecfg_t cfg;
	ng_utp_nodestat_t stat;
} *priv_p;



typedef struct ng_utp_filter_s {
	hook_p		h0; /* Hook for bypass from h1. */
	hook_p		h1; /* Hook for bypass from h0. */
	uint16_t	l3_p;	/* 0 - ethernet, 0x0800 - IPv4. */
	uint16_t	action; /* pass / mark & bypass / drop / send rst. */
	uint32_t	flags; /* Mark L2/L3... */
	uint16_t	label_l2;/* Pass_mark case: L2 - Priority Code Point. */
	uint8_t		label_l3;/* Pass_mark case: L3 ip_tos/Traffic Class code. */
	ng_utp_filterstats_t stat; /* Filter stats. */
} ng_utp_filter_t;



typedef struct ng_utp_hook_private_s {
	hook_p		h_pass; /* Hook for bypass */
	ng_utp_filter_t *filter; /*  */
} ng_utp_hook_private_t, *hpriv_p;



// Netgraph node methods
static ng_constructor_t		ng_utp_constructor;
static ng_shutdown_t		ng_utp_shutdown;
static ng_newhook_t		ng_utp_newhook;
static ng_rcvmsg_t		ng_utp_rcvmsg;
static ng_rcvdata_t		ng_utp_rcvdata;
static ng_disconnect_t		ng_utp_disconnect;



/* Parse type for struct ng_utp_nodecfg_s */
static const struct ng_parse_struct_field ng_utp_nodecfg_type_fields[]
	= NG_UTP_NODE_CFG_TYPE_INFO;
static const struct ng_parse_type ng_utp_nodecfg_type = {
	&ng_parse_struct_type,
	&ng_utp_nodecfg_type_fields
};

/* Parse type for struct ng_utp_nodestatus_s */
static const struct ng_parse_struct_field ng_utp_nodestatus_type_fields[]
	= NG_UTP_NODE_STATUS_TYPE_INFO;
static const struct ng_parse_type ng_utp_nodestatus_type = {
	&ng_parse_struct_type,
	&ng_utp_nodestatus_type_fields
};

/* Parse type for struct ng_utp_nodestat_s */
static const struct ng_parse_struct_field ng_utp_nodestat_type_fields[]
	= NG_UTP_NODE_STAT_TYPE_INFO;
static const struct ng_parse_type ng_utp_nodestat_type = {
	&ng_parse_struct_type,
	&ng_utp_nodestat_type_fields
};


/* Parse type for struct ng_utp_add_filter. */
static const struct ng_parse_struct_field ng_utp_add_filter_fields[] =
	NG_UTP_ADD_FILTER_FIELDS;
static const struct ng_parse_type ng_utp_add_filter_type = {
	&ng_parse_struct_type,
	&ng_utp_add_filter_fields
};

/* Parse type for struct ng_utp_filterstats_s */
static const struct ng_parse_struct_field ng_utp_filterstats_type_fields[]
	= NG_UTP_HOOK_STATS_TYPE_INFO;
static const struct ng_parse_type ng_utp_filterstats_type = {
	&ng_parse_struct_type,
	&ng_utp_filterstats_type_fields
};


/* Parse type for one struct ng_viface_arp_tbl_entry. */
static const struct ng_parse_struct_field ng_utp_conn_tbl_entry4_type_fields[]
	= NG_UTP_CONN_TBL_ENTRY4_TYPE_INFO;
static const struct ng_parse_type ng_utp_conn_tbl_entry4_type = {
	&ng_parse_struct_type,
	&ng_utp_conn_tbl_entry4_type_fields
};


/* Parse type for 'ip addrs' array in struct ng_utp_conn_tbl4_list_s. */
static int
ng_utp_conn_tbl4_ary_getLength(const struct ng_parse_type *type,
			       const u_char *start, const u_char *buf)
{
	const struct ng_utp_conn_tbl4_list_s *al;

	al = (const struct ng_utp_conn_tbl4_list_s *) (buf - 
		offsetof(struct ng_utp_conn_tbl4_list_s, conn_entryes));
	return (al->count);
}

static const struct ng_parse_array_info ng_utp_conn_tbl4_ary_info = {
	&ng_utp_conn_tbl_entry4_type,
	&ng_utp_conn_tbl4_ary_getLength,
	NULL
};
static const struct ng_parse_type ng_utp_conn_tbl4_list_ary_type = {
	&ng_parse_array_type,
	&ng_utp_conn_tbl4_ary_info
};

/* Parse type for struct ng_viface_arp_tbl*/
static const struct ng_parse_struct_field ng_utp_conn_tbl4_list_fields[]
	= NG_UTP_CONN_TBL4_LIST_TYPE_INFO(&ng_utp_conn_tbl4_list_ary_type);
static const struct ng_parse_type ng_utp_conn_tbl4_list_type = {
	&ng_parse_struct_type,
	&ng_utp_conn_tbl4_list_fields
};



/* List of commands and how to convert arguments to/from ASCII  */
static const struct ng_cmdlist ng_utp_cmds[] = {
	{
		NGM_UTP_COOKIE,
		NGM_UTP_GET_NODE_CFG,
		"getcfg",
		NULL,
		&ng_utp_nodecfg_type,
	},
	{
		NGM_UTP_COOKIE,
		NGM_UTP_SET_NODE_CFG,
		"setcfg",
		&ng_utp_nodecfg_type,
		NULL,
	},

	{
		NGM_UTP_COOKIE,
		NGM_UTP_GET_NODE_STATUS,
		"getstatus",
		NULL,
		&ng_utp_nodestatus_type,
	},

	{
		NGM_UTP_COOKIE,
		NGM_UTP_GET_NODE_STAT,
		"getstat",
		NULL,
		&ng_utp_nodestat_type,
	},
	{
		NGM_UTP_COOKIE,
		NGM_UTP_CLR_NODE_STAT,
		"clrstat",
		NULL,
		NULL,
	},
	{
		NGM_UTP_COOKIE,
		NGM_UTP_GETCLR_NODE_STAT,
		"getclrstat",
		NULL,
		&ng_utp_nodestat_type,
	},

	{
		NGM_UTP_COOKIE,
		NGM_UTP_FLUSH_CONN4,
		"flushconn4",
		NULL,
		NULL
	},
	{
		NGM_UTP_COOKIE,
		NGM_UTP_FLUSH_CONN6,
		"flushconn6",
		NULL,
		NULL
	},
	{
		NGM_UTP_COOKIE,
		NGM_UTP_GET_CONN_TABLE4,
		"getconntbl4",
		NULL,
		&ng_utp_conn_tbl4_list_type
	},
	{
		NGM_UTP_COOKIE,
		NGM_UTP_GET_CONN_TABLE6,
		"getconntbl6",
		NULL,
		NULL
	},

	{
		NGM_UTP_COOKIE,
		NGM_UTP_ADD_FILTER,
		"addfilter",
		&ng_utp_add_filter_type,
		NULL
	},
	{
		NGM_UTP_COOKIE,
		NGM_UTP_GET_FILTER,
		"getfilter",
		&ng_parse_hookbuf_type,
		&ng_utp_add_filter_type
	},
	{
		NGM_UTP_COOKIE,
		NGM_UTP_DEL_FILTER,
		"delfilter",
		&ng_parse_hookbuf_type,
		NULL
	},
	{
		NGM_UTP_COOKIE,
		NGM_UTP_GET_FILTER_STATS,
		"getfilterstats",
		&ng_parse_hookbuf_type,
		&ng_utp_filterstats_type,
	},
	{
		NGM_UTP_COOKIE,
		NGM_UTP_CLR_FILTER_STATS,
		"clrfilterstats",
		&ng_parse_hookbuf_type,
		NULL,
	},
	{
		NGM_UTP_COOKIE,
		NGM_UTP_GETCLR_FILTER_STATS,
		"getclrfilterstats",
		&ng_parse_hookbuf_type,
		&ng_utp_filterstats_type,
	},
	{
		NGM_UTP_COOKIE,
		NGM_UTP_LIST_FILTERS,
		"listfilters",
		NULL,
		&ng_utp_add_filter_type
	},

	{ 0 }
};


/* Node type descriptor */
static struct ng_type typestruct = {
	.version =	NG_ABI_VERSION,
	.name =		NG_UTP_NODE_TYPE,
	.constructor =	ng_utp_constructor,
	.rcvmsg =	ng_utp_rcvmsg,
	.shutdown =	ng_utp_shutdown,
	.newhook =	ng_utp_newhook,
	.rcvdata =	ng_utp_rcvdata,
	.disconnect =	ng_utp_disconnect,
	.cmdlist =	ng_utp_cmds,
};
NETGRAPH_INIT(utp, &typestruct);




//************************************************************************
//			HELPER STUFF
//************************************************************************

static __inline int
m_chk(struct mbuf **mp, int len)
{

	if ((*mp)->m_pkthdr.len < len) {
		//m_freem((*mp));
		//(*mp) = NULL;
		return (EINVAL);
	}
	if ((*mp)->m_len < len && ((*mp) = m_pullup((*mp), len)) == NULL)
		return (ENOBUFS);
	
	return (0);
}

/*
 * RFC 1624 Incremental Internet Checksum (May 1994)
 * new_sum = ~(~old_sum + ~old_halfword + new_halfword)
 * HC' = ~(~HC + ~m + m')
 */
static __inline uint16_t
csum_inc(uint16_t csum, uint16_t old16, uint16_t new16)
{
	register uint32_t sum;

	sum = ((~csum) & 0xffff) + ((~old16) & 0xffff) + new16;
	sum = (sum & 0xffff) + (sum >> 16);
	return (~(sum + (sum >> 16)));
}

//************************************************************************
//			Host L2 address STUFF
//************************************************************************

static int
hst_store_init(node_p node, hst_store_p h_store)
{

	bzero(h_store, sizeof(hst_store_t));

	h_store->hst_conn_zn = uma_zcreate("ng_utp ip_conn", sizeof(hst_conn_t),
	    NULL, NULL, NULL, NULL, UMA_ALIGN_CACHE, 0);
	h_store->utp_conn_zn = uma_zcreate("ng_utp utp_conns", sizeof(utp_conn_t),
	    NULL, NULL, NULL, NULL, UMA_ALIGN_CACHE, 0);
	for (int i = 0; i < HASH_BUCKETS_COUNT; i ++) {
		mtx_init(&h_store->buckets[i].rw_lock, "ng_utp_bucket", NULL,
		    MTX_DEF);
		TAILQ_INIT(&h_store->buckets[i].hst_head);
	}
	ng_callout_init(&h_store->timer);
	/* Start timer; timer is always running while node is alive */
	ng_callout(&h_store->timer, node, NULL, hz, hst_store_timer, h_store, 0);

	return (0);
}

static void
hst_store_destroy(node_p node, hst_store_p h_store)
{
	hst_conn_p hst_conn, hst_conn_temp;

	ng_uncallout(&h_store->timer, node);
	
	for (int i = 0; i < HASH_BUCKETS_COUNT; i ++) {
		mtx_lock(&h_store->buckets[i].rw_lock);
		TAILQ_FOREACH_SAFE(hst_conn, &h_store->buckets[i].hst_head, next,
		    hst_conn_temp) {
			mtx_lock(&hst_conn->rw_lock);
			hst_store_conn_free(h_store, &h_store->buckets[i], hst_conn);
		}
		mtx_unlock(&h_store->buckets[i].rw_lock);
		mtx_destroy(&h_store->buckets[i].rw_lock);
	}

	uma_zdestroy(h_store->utp_conn_zn);
	uma_zdestroy(h_store->hst_conn_zn);
	bzero(h_store, sizeof(hst_store_t));
}

static uint32_t 
hst_store_get_bucket_proto_idx(hst_store_p h_store, hst_bucket_p bucket)
{

	/*if (&h_store->buckets[0] > bucket ||
	    &h_store->buckets[HASH_BUCKETS_COUNT] < bucket)
		return (-1);*/

	return (((bucket - &h_store->buckets[0]) / HASH_BUCKETS));
}

static hst_bucket_p 
hst_store_proto_buckets(hst_store_p h_store, uint16_t l3_p)
{

	switch (l3_p) {
	case ETHERTYPE_IP:
		return (&h_store->buckets[0]);
		break;
	case ETHERTYPE_IPV6:
		return (&h_store->buckets[(HASH_BUCKETS * 1)]);
		break;
	default:
		printf("%s: error: unknown ether type = %u\n", __func__, l3_p);
	}

	return (NULL);
}

static hst_bucket_p 
hst_store_get_bucket(hst_store_p h_store, uint16_t l3_p, u_char *saddr,
    u_char *daddr, uint16_t sport, uint16_t dport, uint32_t *addr_len,
    uint32_t *proto_idx)
{
	hst_bucket_p bucket = NULL, buckets;
	uint32_t ret_addr_len = 0, ret_proto_idx = 0;

	/* Get buckets for L3 proto */
	buckets = hst_store_proto_buckets(h_store, l3_p);
	if (NULL == buckets)
		return (NULL);

	switch (l3_p) {
	case ETHERTYPE_IP:
		bucket = &buckets[HASH_IPV4_HST(saddr, daddr, &sport, &dport)];
		ret_addr_len = sizeof(struct in_addr);
		ret_proto_idx = 0;
		break;
	case ETHERTYPE_IPV6:
		bucket = &buckets[HASH_IPV6_HST(saddr, daddr, &sport, &dport)];
		ret_addr_len = sizeof(struct in6_addr);
		ret_proto_idx = 1;
		break;
	}

	if (NULL != addr_len)
		(*addr_len) = ret_addr_len;
	if (NULL != proto_idx)
		(*proto_idx) = ret_proto_idx;

	return (bucket);
}

static hst_conn_p 
hst_store_conn_alloc(hst_store_p h_store, uint16_t l3_p, u_char *saddr,
    u_char *daddr, uint16_t sport, uint16_t dport)
{
	hst_bucket_p bucket;
	hst_conn_p hst_conn;
	uint32_t addr_len, proto_idx;

	hst_conn = uma_zalloc(h_store->hst_conn_zn, (M_NOWAIT | M_ZERO));
	if (NULL == hst_conn)
		return (hst_conn);

	/* Get buckets for L3 proto */
	bucket = hst_store_get_bucket(h_store, l3_p, saddr, daddr, sport, dport,
	    &addr_len, &proto_idx);
	if (NULL == bucket)
		return (NULL);
	bcopy(saddr, &hst_conn->saddr, addr_len);
	bcopy(daddr, &hst_conn->daddr, addr_len);
	hst_conn->sport = sport;
	hst_conn->dport = dport;
	//hst_conn->lock_wait_count = 0;
	hst_conn->create_time = time_second;
	mtx_init(&hst_conn->rw_lock, "ng_utp_conn", NULL, (MTX_DEF | MTX_RECURSE));
	mtx_lock(&hst_conn->rw_lock);

	/* add to bucket */
	mtx_lock(&bucket->rw_lock);
	TAILQ_INSERT_HEAD(&bucket->hst_head, hst_conn, next);
	mtx_unlock(&bucket->rw_lock);
	h_store->hst_conns[proto_idx] ++;

	return (hst_conn);
}

static void
hst_store_conn_free(hst_store_p h_store, hst_bucket_p bucket, hst_conn_p hst_conn)
{
	utp_conn_p utp_conn, utp_conn_temp;

	/* 
	 * XXX bucket must be locked before this func called, or no guaratis that
	 * hst item exist - pointer point to free space
	 */
	//if (mtx_owned(&bucket->rw_lock) == 0)
	//	panic("hst_store_conn_free: bucket not locked by calling thread");
	h_store->hst_conns[hst_store_get_bucket_proto_idx(h_store, bucket)] --;
	TAILQ_REMOVE(&bucket->hst_head, hst_conn, next);

	/* wait untill all threads do their job with item */
	while (0 == atomic_cmpset_32(&hst_conn->lock_wait_count, 0, 0)) {
		mtx_unlock(&hst_conn->rw_lock);
		//XXX sleep(100);
		mtx_lock(&hst_conn->rw_lock);
	}
	TAILQ_FOREACH_SAFE(utp_conn, &hst_conn->conns_head, next, utp_conn_temp) {
		hst_store_utp_conn_free(h_store, hst_conn, utp_conn);
	}
	mtx_unlock(&hst_conn->rw_lock);
	mtx_destroy(&hst_conn->rw_lock);
	uma_zfree(h_store->hst_conn_zn, hst_conn);
}

static utp_conn_p 
hst_store_utp_conn_alloc(hst_store_p h_store, hst_conn_p hst_conn, int hst_rev_state,
    int pkt_ver, uint32_t connid)
{
	utp_conn_p utp_conn;

	utp_conn = uma_zalloc(h_store->utp_conn_zn, (M_NOWAIT | M_ZERO));
	if (NULL == utp_conn)
		return (utp_conn);

	utp_conn->connid = connid;
	utp_conn->pkt_ver = pkt_ver;
	utp_conn->hst_rev_state = hst_rev_state;
	//utp_conn->req_last = hst_conn->last_update = time_second;
	TAILQ_INSERT_HEAD(&hst_conn->conns_head, utp_conn, next);
	h_store->utp_conns[pkt_ver] ++;
	hst_conn->conns_count ++;

	return (utp_conn);
}

static void
hst_store_utp_conn_free(hst_store_p h_store, hst_conn_p hst_conn,
    utp_conn_p utp_conn)
{
	/* XXX hst_conn must be locked before this func called */
	//if (mtx_owned(&hst_conn->rw_lock) == 0)
	//	panic("hst_store_utp_conn_free: host not locked by calling thread");
	hst_conn->conns_count --;
	h_store->utp_conns[utp_conn->pkt_ver] --;
	TAILQ_REMOVE(&hst_conn->conns_head, utp_conn, next);
	uma_zfree(h_store->utp_conn_zn, utp_conn);
}

static void
hst_store_timer(node_p node, hook_p hook, void *arg1, int arg2)
{
	const priv_p priv = NG_NODE_PRIVATE(node);
	hst_store_p h_store = arg1;
	hst_bucket_p bucket;
	hst_conn_p hst_conn, hst_conn_temp;
	utp_conn_p utp_conn, utp_conn_temp;
	time_t time_now;

	time_now = time_second;
	for(int i = 0; i < HASH_BUCKETS_COUNT; i++) {
		bucket = &h_store->buckets[i];
		mtx_lock(&bucket->rw_lock);
		TAILQ_FOREACH_SAFE(hst_conn, &bucket->hst_head, next, hst_conn_temp) {
			/* is session dead? */
			if (0 == atomic_cmpset_32(&hst_conn->lock_wait_count, 0, 0))
				continue; /* Item steel used by some one else. */
			if (0 == mtx_trylock(&hst_conn->rw_lock))
				continue; /* Item steel used by some one else. */

			/* Host item LOCKED! */
			if ((time_now - hst_conn->last_update) >
			    priv->cfg.conn_idle_timeout) {
				/* Delete host with all connections by timeout. */
				hst_store_conn_free(h_store, bucket, hst_conn);
				continue;
			}

			TAILQ_FOREACH_SAFE(utp_conn, &hst_conn->conns_head, next,
			    utp_conn_temp) {
				if ((time_now - max(utp_conn->req_last,
				    utp_conn->repl_last)) >
				    priv->cfg.conn_idle_timeout)
					/* delete connection by timeout */
					hst_store_utp_conn_free(h_store, hst_conn,
					    utp_conn);
			}
			if (0 == hst_conn->conns_count) {
				/* Destroy host state with no connections. */
				hst_store_conn_free(h_store, bucket, hst_conn);
				continue;
			}
			mtx_unlock(&hst_conn->rw_lock);
		}
		mtx_unlock(&bucket->rw_lock);
	}

	/* Register a new timeout, keeping the existing node reference. */
	ng_callout(&h_store->timer, node, NULL, hz, hst_store_timer, h_store, 0);
}

static void 
hst_store_flush_conn(hst_store_p h_store, uint16_t l3_p)
{
	hst_bucket_p buckets, bucket;
	hst_conn_p hst_conn, hst_conn_temp;

	/* Get buckets for L3 proto */
	buckets = hst_store_proto_buckets(h_store, l3_p);
	if (NULL == buckets)
		return;

	for (int i = 0; i < HASH_BUCKETS; i ++) {
		bucket = &buckets[i];
		mtx_lock(&bucket->rw_lock);
		TAILQ_FOREACH_SAFE(hst_conn, &bucket->hst_head, next, hst_conn_temp) {
			mtx_lock(&hst_conn->rw_lock);
			hst_store_conn_free(h_store, bucket, hst_conn);
		}
		mtx_unlock(&bucket->rw_lock);
	}
}

static hst_conn_p 
hst_store_find_hst_conn(hst_store_p h_store, uint16_t l3_p, u_char *saddr,
    u_char *daddr, uint16_t sport, uint16_t dport, int *rev)
{
	hst_bucket_p bucket;
	hst_conn_p hst_conn;
	uint32_t addr_len;

	/* Get buckets for L3 proto */
	bucket = hst_store_get_bucket(h_store, l3_p, saddr, daddr, sport, dport,
	    &addr_len, NULL);
	if (NULL == bucket)
		return (NULL);

	mtx_lock(&bucket->rw_lock);
	TAILQ_FOREACH(hst_conn, &bucket->hst_head, next) {
		if (hst_conn->sport == sport && hst_conn->dport == dport &&
		    0 == bcmp(saddr, hst_conn->saddr, addr_len) &&
		    0 == bcmp(daddr, hst_conn->daddr, addr_len)) {
			/*
			 * Mark item as pending to use to prevent destroy by timeout
			 * release bucket and start locking host item.
			 */
			atomic_add_32(&hst_conn->lock_wait_count, 1);
			mtx_unlock(&bucket->rw_lock);
			mtx_lock(&hst_conn->rw_lock);
			atomic_subtract_32(&hst_conn->lock_wait_count, 1);
			(*rev) = 0;
			return (hst_conn);
		}
		/* Reverce check. */
		if (hst_conn->sport == dport && hst_conn->dport == sport &&
		    0 == bcmp(daddr, hst_conn->saddr, addr_len) &&
		    0 == bcmp(saddr, hst_conn->daddr, addr_len)) {
			/*
			 * Mark item as pending to use to prevent destroy by timeout
			 * release bucket and start locking host item.
			 */
			atomic_add_32(&hst_conn->lock_wait_count, 1);
			mtx_unlock(&bucket->rw_lock);
			mtx_lock(&hst_conn->rw_lock);
			atomic_subtract_32(&hst_conn->lock_wait_count, 1);
			(*rev) = 1;
			return (hst_conn);
		}
	}
	mtx_unlock(&bucket->rw_lock);

	return (NULL);
}

static utp_conn_p 
hst_store_find_utp_conn(hst_conn_p hst_conn, int hst_rev_state, int pkt_ver,
    uint32_t connid, int *rev)
{
	utp_conn_p utp_conn;

	TAILQ_FOREACH(utp_conn, &hst_conn->conns_head, next) {
		if (pkt_ver != utp_conn->pkt_ver) /* Packet ver check. */
			continue;

		if ((utp_conn->connid == connid &&	/* Conn direction check. */
		    utp_conn->hst_rev_state == hst_rev_state) ||
		    (utp_conn->connid == (connid + 1) &&
		    utp_conn->hst_rev_state != hst_rev_state)) {
			/*
			 * req_SYN and rep_STATE have same ConnID,
			 * but all next req_* have ConnID+1, rep_* allways
			 * have ConnID from first req_SYN packet.
			 */
			(*rev) = (utp_conn->connid - connid);
			return (utp_conn);
		} else if (utp_conn->connid == (connid - 1) &&
		    utp_conn->hst_rev_state != hst_rev_state &&
		    0 == utp_conn->repl_count &&
		    (utp_conn->req_state & CONN_F_ST_SYN) == 0) {
			/* First packet, that create utp_conn was reply, fix it. */
			/*printf("%s: Converted conn: %u -> %u, rev_state: %i -> %i "
			    "pkt_ver = %i\n", __func__,
			    utp_conn->connid, (utp_conn->connid + 1),
			    utp_conn->hst_rev_state, hst_rev_state,
			    pkt_ver);//*/
			utp_conn->hst_rev_state = hst_rev_state;
			utp_conn->connid ++;
			utp_conn->repl_last = utp_conn->req_last;
			utp_conn->repl_count = utp_conn->req_count;
			utp_conn->repl_state = utp_conn->req_state;
			utp_conn->req_last = 0;
			utp_conn->req_count = 0;
			utp_conn->req_state = 0;

			(*rev) = 0;
			return (utp_conn);
		}
	}

	return (NULL);
}

/*
 * Dump all IPv4 to buffer
 */
static int
hst_store_entryes4(hst_store_p h_store, u_char *buff, uint32_t buff_size,
    uint32_t *entryes_cnt)
{
	int error = 0;
	hst_bucket_p bucket, buckets;
	hst_conn_p hst_conn;
	ng_utp_conn_tbl_entry4_t *entry4;
	uint32_t i, entry_size, entryes_cnt_local;
	time_t time_now;

	if (NULL == buff || 0 == buff_size)
		return (EINVAL);

	/* Get buckets for L3 proto */
	time_now = time_second;
	entryes_cnt_local = 0;
	entry_size = sizeof(ng_utp_conn_tbl_entry4_t);
	buckets = hst_store_proto_buckets(h_store, ETHERTYPE_IP);

	for(i = 0; i < HASH_BUCKETS; i ++) {
		bucket = &buckets[i];
		mtx_lock(&bucket->rw_lock);
		TAILQ_FOREACH(hst_conn, &bucket->hst_head, next) {
			if (buff_size < entry_size) {
				error = ENOMEM;
				break;// XXX
			}

			entry4 = (ng_utp_conn_tbl_entry4_t *)buff;
			bcopy(hst_conn->saddr, entry4->saddr, 4);
			bcopy(hst_conn->daddr, entry4->daddr, 4);
			entry4->sport = ntohs(hst_conn->sport);
			entry4->dport = ntohs(hst_conn->dport);
			entry4->conns_count = hst_conn->conns_count;
			entry4->lifetime = (time_now - hst_conn->create_time);
			entry4->idle = (time_now - hst_conn->last_update);

			buff += entry_size;
			buff_size -= entry_size;
			entryes_cnt_local ++;
		}
		mtx_unlock(&bucket->rw_lock);
	}
	if (NULL != entryes_cnt)
		(*entryes_cnt) = entryes_cnt_local;

	return (error);
}


//************************************************************************
//			NETGRAPH NODE STUFF
//************************************************************************

// Node constructor
static int
ng_utp_constructor(node_p node)
{
	priv_p priv;

	// Allocate private structure
	priv = malloc(sizeof(*priv), M_NETGRAPH_UTP, (M_NOWAIT | M_ZERO));
	if (NULL == priv)
		return (ENOMEM);

	hst_store_init(node, &priv->h_store);
	/* Apply defaults. */
	uma_zone_set_max(priv->h_store.hst_conn_zn, NG_UTP_MAX_HOST_CONNS);
	uma_zone_set_max(priv->h_store.utp_conn_zn, NG_UTP_MAX_UTP_CONNS);
	priv->cfg.conn_ack_timeout = NG_UTP_CONN_ACK_TIMEOUT;
	priv->cfg.conn_idle_timeout = NG_UTP_CONN_IDLE_TIMEOUT;

	NG_NODE_SET_PRIVATE(node, priv);

	return (0);
}

// Nuke node
static int
ng_utp_shutdown(node_p node)
{
	const priv_p priv = NG_NODE_PRIVATE(node);

	// Take down netgraph node
	hst_store_destroy(node, &priv->h_store);
	bzero(priv, sizeof(*priv));
	free(priv, M_NETGRAPH_UTP);
	NG_NODE_SET_PRIVATE(node, NULL);
	NG_NODE_UNREF(node);

	return (0);
}

// Give our ok for a hook to be added
static int
ng_utp_newhook(node_p node, hook_p hook, const char *name)
{

	NG_HOOK_SET_PRIVATE(hook, NULL);

	return (0);
}

static void
ng_utp_filter_delete(ng_utp_filter_t *filter)
{
	hpriv_p hpriv;

	if (NULL == filter)
		return;

	if (NULL != filter->h0) {
		hpriv = NG_HOOK_PRIVATE(filter->h0);
		NG_HOOK_SET_PRIVATE(filter->h0, NULL);
		bzero(hpriv, sizeof(ng_utp_hook_private_t));
		free(hpriv, M_NETGRAPH_UTP);
	}
	if (NULL != filter->h1 && filter->h0 != filter->h1) {
		hpriv = NG_HOOK_PRIVATE(filter->h1);
		NG_HOOK_SET_PRIVATE(filter->h1, NULL);
		bzero(hpriv, sizeof(ng_utp_hook_private_t));
		free(hpriv, M_NETGRAPH_UTP);
	}
	bzero(filter, sizeof(ng_utp_filter_t));
	free(filter, M_NETGRAPH_UTP);
}

// Hook disconnection
static int
ng_utp_disconnect(hook_p hook)
{
	const node_p node = NG_HOOK_NODE(hook);
	//const priv_p priv = NG_NODE_PRIVATE(node);

	if ((0 == NG_NODE_NUMHOOKS(node)) && (NG_NODE_IS_VALID(node))) {
		ng_rmnode_self(node);
	} else if (NULL != NG_HOOK_PRIVATE(hook)) {
		ng_utp_filter_delete(((hpriv_p)NG_HOOK_PRIVATE(hook))->filter);
	}

	return (0);
}

// Receive a control message.
static int
ng_utp_rcvmsg(node_p node, item_p item, hook_p lasthook)
{
	int error = 0;
	const priv_p priv = NG_NODE_PRIVATE(node);
	struct ng_mesg *msg;
	struct ng_mesg *resp = NULL;

	NGI_GET_MSG(item, msg);

	if (NGM_UTP_COOKIE != msg->header.typecookie) {
		error = EINVAL;
		goto respond_msg;
	}

	switch (msg->header.cmd) {
	case NGM_UTP_GET_NODE_CFG: /* Get node config. */
		NG_MKRESPONSE(resp, msg, sizeof(ng_utp_nodecfg_t), M_NOWAIT);
		if (NULL == resp) {
			error = ENOMEM;
			break;
		}
		bcopy(&priv->cfg, resp->data, sizeof(ng_utp_nodecfg_t));
		break;
	case NGM_UTP_SET_NODE_CFG: /* Set node config. */
	    {
		ng_utp_nodecfg_t *n_cfg;

		/* Check that message is long enough. */
		if (sizeof(ng_utp_nodecfg_t) != msg->header.arglen ||
		    NULL == msg->data) {
			error = EINVAL;
			break;
		}
		n_cfg = (ng_utp_nodecfg_t *)msg->data;
		uma_zone_set_max(priv->h_store.hst_conn_zn, n_cfg->ip_conns_max);
		uma_zone_set_max(priv->h_store.utp_conn_zn, n_cfg->utp_conns_max);
		priv->cfg.ip_conns_max = n_cfg->ip_conns_max;
		priv->cfg.utp_conns_max = n_cfg->utp_conns_max;
		if (0 != n_cfg->conn_ack_timeout)
			priv->cfg.conn_ack_timeout = n_cfg->conn_ack_timeout;
		if (0 != n_cfg->conn_idle_timeout)
			priv->cfg.conn_idle_timeout = n_cfg->conn_idle_timeout;
		break;
	    }
	case NGM_UTP_GET_NODE_STATUS: /* Get node status. */
	    {
		ng_utp_nodestatus_t *n_st;

		NG_MKRESPONSE(resp, msg, sizeof(ng_utp_nodestatus_t), M_NOWAIT);
		if (NULL == resp) {
			error = ENOMEM;
			break;
		}
		n_st = (ng_utp_nodestatus_t *)resp->data;
		n_st->ipv4_conns = priv->h_store.hst_conns[0];
		n_st->ipv6_conns = priv->h_store.hst_conns[1];
		n_st->ip_conns_total = uma_zone_get_cur(priv->h_store.hst_conn_zn);
		n_st->ip_conns_max = uma_zone_get_max(priv->h_store.hst_conn_zn);
		n_st->utpv0_conns = priv->h_store.utp_conns[0];
		n_st->utpv1_conns = priv->h_store.utp_conns[1];
		n_st->utp_conns_total = uma_zone_get_cur(priv->h_store.utp_conn_zn);
		n_st->utp_conns_max = uma_zone_get_max(priv->h_store.utp_conn_zn);
		n_st->mem_ip_conns = (n_st->ip_conns_total * sizeof(hst_conn_t));
		n_st->mem_utp_conns = (n_st->utp_conns_total * sizeof(utp_conn_t));
		n_st->mem_total = (n_st->mem_ip_conns + n_st->mem_utp_conns);

		KASSERT((n_st->ipv4_conns + n_st->ipv6_conns) ==
		    n_st->utp_conns_total,
		    ("%s: NGM_UTP_GET_NODE_STATUS:"
		    " ipv4_conns + ipv6_conns != utp_conns_total\n", __func__));

		KASSERT((n_st->utpv0_conns + n_st->utpv1_conns) ==
		    n_st->utp_conns_total,
		    ("%s: NGM_UTP_GET_NODE_STATUS:"
		    " utpv0_conns + utpv1_conns != utp_conns_total\n", __func__));
		break;
	    }
	case NGM_UTP_GET_NODE_STAT:/* Get/clr node statistics. */
	case NGM_UTP_CLR_NODE_STAT:
	case NGM_UTP_GETCLR_NODE_STAT:
		/* Get/clear stat */
		if (NGM_UTP_CLR_NODE_STAT != msg->header.cmd) {
			ng_utp_nodestat_t *stat;

			NG_MKRESPONSE(resp, msg, sizeof(ng_utp_nodestat_t),
				      M_NOWAIT);
			if (NULL == resp) {
				error = ENOMEM;
				break;
			}
			stat = (ng_utp_nodestat_t *)resp->data;
			bcopy(&priv->stat, resp->data, sizeof(ng_utp_nodestat_t));
			stat->ip_conns_total = (stat->ipv4_conns + stat->ipv6_conns);
			stat->utp_conns_total = (stat->utpv0_conns +
			    stat->utpv1_conns);
		}
		if (NGM_UTP_GET_NODE_STAT != msg->header.cmd)
			bzero(&priv->stat, sizeof(ng_utp_nodestat_t));
		break;
	case NGM_UTP_FLUSH_CONN4:
		hst_store_flush_conn(&priv->h_store, ETHERTYPE_IP);
		break;
	case NGM_UTP_FLUSH_CONN6:
		hst_store_flush_conn(&priv->h_store, ETHERTYPE_IPV6);
		break;
	case NGM_UTP_GET_CONN_TABLE4:
	    {
		uint32_t count, buff_size;
		ng_utp_conn_tbl4_list_t *conn_list4;

		count = priv->h_store.hst_conns[0];
		buff_size = ((count * sizeof(ng_utp_conn_tbl_entry4_t)) +
		    sizeof(ng_utp_conn_tbl4_list_t) + 1024);
		NG_MKRESPONSE(resp, msg, buff_size, M_NOWAIT);
		if (NULL == resp) {
			error = ENOMEM;
			break;
		}
		conn_list4 = (ng_utp_conn_tbl4_list_t *)resp->data;
		error = hst_store_entryes4(&priv->h_store,
		    (u_char *)conn_list4->conn_entryes,
		    (buff_size - sizeof(ng_utp_conn_tbl4_list_t)), &count);
		conn_list4->count = count;
		break;
	    }
	case NGM_UTP_GET_CONN_TABLE6: /* Not implemented. */
		error = EINVAL;
		break;
	case NGM_UTP_ADD_FILTER:
	    {
		hook_p h0, h1;
		hpriv_p hpriv;
		ng_utp_filter_t *filter;
		ng_utp_add_filter_t *add_filter;

		/* Check that message is long enough. */
		if (sizeof(ng_utp_add_filter_t) != msg->header.arglen) {
			error = EINVAL;
			break;
		}
		add_filter = (ng_utp_add_filter_t *)msg->data;
		/* Sanity check the l3_p value. */
		if (0 != add_filter->l3_proto &&
		    ETHERTYPE_IP != add_filter->l3_proto &&
		    ETHERTYPE_IPV6 != add_filter->l3_proto) {
			error = EINVAL;
			break;
		}

		switch (add_filter->action) {
		case UTP_FLT_ACCTION_PASS:
		case UTP_FLT_ACCTION_DROP:
		case UTP_FLT_ACCTION_SND_RST:
			break;
		case UTP_FLT_ACCTION_PASS_MARK:
			if (0 != add_filter->l3_proto) /* No L2. */
				add_filter->flags &= ~UTP_FLT_FLAG_MARK_L2;
			/* Check additional params. */
			if (0 == (add_filter->flags &
			    (UTP_FLT_FLAG_MARK_L2 | UTP_FLT_FLAG_MARK_L3)))
				add_filter->action = UTP_FLT_ACCTION_PASS;
			else if (add_filter->label_l2 & ~0x07 ||
			    add_filter->label_l3 & ~0x3f)
				error = EINVAL;
			break;
		default:
			error = EINVAL;
		}
		if (0 != error)
			break;

		/* Check that a referenced hook exists. */
		h0 = ng_findhook(node, add_filter->h0);
		h1 = ng_findhook(node, add_filter->h1);
		if (NULL == h0 && NULL == h1) {
			error = ENOENT;
			break;
		}
		/* And is not already in service. */
		if ((NULL != h0 && NULL != NG_HOOK_PRIVATE(h0)) ||
		    (NULL != h1 && NULL != NG_HOOK_PRIVATE(h1))) {
			error = EEXIST;
			break;
		}
		/* Create filter. */
		filter = malloc(sizeof(ng_utp_filter_t), M_NETGRAPH_UTP,
		    (M_NOWAIT | M_ZERO));
		if (NULL == filter) {
			error = ENOMEM;
			break;
		}
		/* Link filter and hook together. */
		filter->h0 = h0;
		filter->h1 = h1;
		filter->l3_p = add_filter->l3_proto;
		filter->action = add_filter->action;
		filter->flags = add_filter->flags;
		filter->label_l2 = (add_filter->label_l2 << 13);
		filter->label_l3 = (add_filter->label_l3 << 2);

		if (NULL != h0) {
			hpriv = malloc(sizeof(ng_utp_hook_private_t), M_NETGRAPH_UTP,
			    (M_NOWAIT | M_ZERO));
			if (NULL == hpriv) {
				ng_utp_filter_delete(filter);
				error = ENOMEM;
				break;
			}
			hpriv->h_pass = ((NULL != h1) ? h1 : h0);
			hpriv->filter = filter;
			if (filter->flags & UTP_FLT_FLAG_FORCE_QUEUE_H0)
				NG_HOOK_FORCE_QUEUE(h0);
			NG_HOOK_SET_PRIVATE(h0, hpriv);
		}
		if (NULL != h1) {
			hpriv = malloc(sizeof(ng_utp_hook_private_t), M_NETGRAPH_UTP,
			    (M_NOWAIT | M_ZERO));
			if (NULL == hpriv) {
				ng_utp_filter_delete(filter);
				error = ENOMEM;
				break;
			}
			hpriv->h_pass = ((NULL != h0) ? h0 : h1);
			hpriv->filter = filter;
			if (filter->flags & UTP_FLT_FLAG_FORCE_QUEUE_H1)
				NG_HOOK_FORCE_QUEUE(h1);
			NG_HOOK_SET_PRIVATE(h1, hpriv);
		}
		break;
	    }
	case NGM_UTP_GET_FILTER: /* Get filter settings. */
	    {
		hook_p hook;
		ng_utp_filter_t *filter;
		ng_utp_add_filter_t *add_filter;

		/* Check that message is long enough. */
		if (NG_HOOKSIZ != msg->header.arglen || NULL == msg->data) {
			error = EINVAL;
			break;
		}
		/* Check that hook exists and is active. */
		hook = ng_findhook(node, (char *)msg->data);
		if (NULL == hook || NULL == NG_HOOK_PRIVATE(hook)) {
			error = ENOENT;
			break;
		}
		/* Allocate buff and return data */
		NG_MKRESPONSE(resp, msg, sizeof(ng_utp_add_filter_t), M_NOWAIT);
		if (NULL == resp) {
			error = ENOMEM;
			break;
		}
		add_filter = (ng_utp_add_filter_t *)resp->data;
		filter = ((hpriv_p)NG_HOOK_PRIVATE(hook))->filter;

		add_filter->l3_proto = filter->l3_p;
		add_filter->action = filter->action;
		add_filter->flags = filter->flags;
		add_filter->label_l2 = (filter->label_l2 >> 13);
		add_filter->label_l3 = (filter->label_l3 >> 2);
		if (NULL != filter->h0) {
			strncpy(add_filter->h0, NG_HOOK_NAME(filter->h0),
			    NG_HOOKSIZ);
		} else {
			bzero(add_filter->h0, NG_HOOKSIZ);
		}
		if (NULL != filter->h1) {
			strncpy(add_filter->h1, NG_HOOK_NAME(filter->h1),
			    NG_HOOKSIZ);
		} else {
			bzero(add_filter->h1, NG_HOOKSIZ);
		}

		break;
	    }
	case NGM_UTP_DEL_FILTER: /* Delete filter. */
	    {
		hook_p hook;

		/* Check that message is long enough. */
		if (NG_HOOKSIZ != msg->header.arglen) {
			error = EINVAL;
			break;
		}
		/* Check that hook exists and is active. */
		hook = ng_findhook(node, (char *)msg->data);
		if (NULL == hook || NULL == NG_HOOK_PRIVATE(hook)) {
			error = ENOENT;
			break;
		}
		ng_utp_filter_delete(((hpriv_p)NG_HOOK_PRIVATE(hook))->filter);

		break;
	    }
	case NGM_UTP_GET_FILTER_STATS: /* Get/clear filter statistics. */
	case NGM_UTP_CLR_FILTER_STATS:
	case NGM_UTP_GETCLR_FILTER_STATS:
	    {
		hook_p hook;
		ng_utp_filterstats_t *stats;

		/* Check that message is long enough. */
		if (NG_HOOKSIZ != msg->header.arglen) {
			error = EINVAL;
			break;
		}
		/* Check that hook exists and is active. */
		hook = ng_findhook(node, (char *)msg->data);
		if (NULL == hook || NULL == NG_HOOK_PRIVATE(hook)) {
			error = ENOENT;
			break;
		}

		stats = &((hpriv_p)NG_HOOK_PRIVATE(hook))->filter->stat;
		if (NGM_UTP_CLR_FILTER_STATS != msg->header.cmd) {
			NG_MKRESPONSE(resp, msg, sizeof(ng_utp_filterstats_t),
			    M_NOWAIT);
			if (NULL == resp) {
				error = ENOMEM;
				break;
			}
			bcopy(stats, resp->data, sizeof(ng_utp_filterstats_t));
		}
		if (NGM_UTP_GET_FILTER_STATS != msg->header.cmd)
			bzero(stats, sizeof(ng_utp_filterstats_t));
		break;
	    }
	case NGM_UTP_LIST_FILTERS:
	
		break;
	default:
		error = EINVAL;
		break;
	}

respond_msg:
	NG_RESPOND_MSG(error, node, item, resp);
	NG_FREE_MSG(msg);

	return (error);
}




#define	M_GET_WRITABLE_SIZE(m)						\
	((m)->m_flags & M_EXT ?						\
	    (M_WRITABLE(m) ? (m)->m_ext.ext_size : 0) :			\
	    ((m)->m_flags & M_PKTHDR ? MHLEN : MLEN) )

#define	M_GET_FREE_SIZE(m)						\
	((m)->m_flags & M_EXT ?						\
	    (M_WRITABLE(m) ? ((m)->m_ext.ext_size - (m)->m_len) : 0) :	\
	    (((m)->m_flags & M_PKTHDR ? MHLEN : MLEN) - (m)->m_len) )


/*
 * Rearange an mbuf chain so that len bytes are contiguous
 * and in the data area of an mbuf (so that mtod and dtom
 * will work for a structure of size len).  Returns the resulting
 * mbuf chain on success, frees it and returns null on failure.
 * If there is room, it will add up to max_protohdr-len extra bytes to the
 * contiguous region in an attempt to avoid being called next time.
 */

#if 0
int
m_pullup_safe(struct mbuf *n, int len)
{
	struct mbuf *m;
	int count;
	int space;
	int free_space, next_free_space;


	if (len > n->m_pkthdr.len)
		len = n->m_pkthdr.len;
	if (n->m_len >= len)
		return (len);
	if (NULL == n->m_next)
		return (n->m_len);

	free_space = M_GET_FREE_SIZE(n);
	next_free_space = M_GET_FREE_SIZE(n->m_next);
	/*
	 * If first mbuf has no cluster, and has room for len bytes
	 * without shifting current data, pullup into it,
	 * otherwise allocate a new mbuf to prepend to the chain.
	 */
	if ((n->m_flags & M_EXT) == 0 &&
	    n->m_data + len < &n->m_dat[MLEN] && n->m_next) {
		if (n->m_len >= len)
			return (n);
		m = n;
		n = n->m_next;
		len -= m->m_len;
	} else {
		if (len > MHLEN)
			goto bad;
		MGET(m, M_DONTWAIT, n->m_type);
		if (m == NULL)
			goto bad;
		m->m_len = 0;
		if (n->m_flags & M_PKTHDR)
			M_MOVE_PKTHDR(m, n);
	}
	space = &m->m_dat[MLEN] - (m->m_data + m->m_len);
	do {
		count = min(min(max(len, max_protohdr), space), n->m_len);
		bcopy(mtod(n, caddr_t), mtod(m, caddr_t) + m->m_len,
		  (uint)count);
		len -= count;
		m->m_len += count;
		n->m_len -= count;
		space -= count;
		if (n->m_len)
			n->m_data += count;
		else
			n = m_free(n);
	} while (len > 0 && n);
	if (len > 0) {
		(void) m_free(m);
		goto bad;
	}
	m->m_next = n;
	return (m);
bad:
	m_freem(n);
	mbstat.m_mpfail++;	/* XXX: No consistency. */
	return (NULL);
}
#endif

// Receive data, and do something with it.
static int
ng_utp_rcvdata(hook_p hook, item_p item)
{
	int error = 0;
	const node_p node = NG_HOOK_NODE(hook);
	const priv_p priv = NG_NODE_PRIVATE(node);
	const hpriv_p hpriv = NG_HOOK_PRIVATE(hook);
	struct mbuf *m;
	u_short l3_p;
	hst_conn_p hst_conn;
	utp_conn_p utp_conn;
	ng_utp_filter_t *filter;
	u_char *saddr, *daddr;
	struct ip *ip;
	struct udphdr *udp;
	u_char *udp_data;
	uint32_t pullup_len = 0, l2_hlen = 0, l3_len, l3_hlen, l4_len;
	int pkt_ver, alloc_conn, do_action;
	uint32_t connid, pkt_type;
	uint16_t seq_nr, ack_nr;
	int hst_rev_state, utp_rev_state;
	time_t time_now;


	NGI_GET_M(item, m);

	if (NULL == hpriv || NULL == hpriv->h_pass) { /* Drop */
		m_freem(m);
		NG_FREE_ITEM(item);
		return (ENETDOWN);
	}

	filter = hpriv->filter;
	if (m->m_flags & (M_BCAST | M_MCAST)) /* XXX | M_PROMISC */
		goto bypass;
	/* Do checks to verify assumptions made by code past this point. */
	if (0 == (m->m_flags & M_PKTHDR)) {
		filter->stat.recvErrors ++;
		error = EINVAL;
		goto drop;
	}

	filter->stat.Octets += m->m_pkthdr.len;
	filter->stat.Packets ++;

	l3_p = filter->l3_p;
	/* LAYER 2 */
	if (0 == l3_p) { /* Ethernet (L2) packet. */
		/* Pool up ethernet header. */
		pullup_len = ETHER_HDR_LEN;
		if (0 != (error = m_chk(&m, pullup_len)))
			goto mchk_err;
		/* Bypass L2 multicast. */
		if (ETHER_IS_MULTICAST(mtod(m, struct ether_header *)->ether_dhost))
			goto bypass;
		l3_p = ntohs((mtod(m, struct ether_header *))->ether_type);
		if (ETHERTYPE_VLAN == l3_p) {
			/* Pool up ethernet - vlan header. */
			pullup_len += ETHER_VLAN_ENCAP_LEN;
			if (0 != (error = m_chk(&m, pullup_len)))
				goto mchk_err;
			l3_p = ntohs(mtod(m, struct ether_vlan_header *)->evl_proto);
		}
		l2_hlen = pullup_len;
	}

	/* LAYER 3 */
	switch (l3_p) {
	case ETHERTYPE_IP:
		filter->stat.IPv4_Packets ++;
		/* Pool up ip header. */
		pullup_len += sizeof(struct ip);
		if (0 != (error = m_chk(&m, pullup_len)))
			goto mchk_err;
		ip = (struct ip *)(mtod(m, u_char *) + l2_hlen);
		l3_len = ntohs(ip->ip_len);
		l3_hlen = (ip->ip_hl << 2);

		/* Basic packet checks, from ip_input, only if hook is L2. */
		if (0 != l2_hlen && (
		    IPVERSION != ip->ip_v ||
		    // 0 < ip->ip_ttl ||
		    sizeof(struct ip) > l3_hlen ||
		    l3_len < l3_hlen ||
		    (l2_hlen + l3_len) > m->m_pkthdr.len ||
		    IN_LOOPBACK(ntohl(ip->ip_dst.s_addr)) ||
		    IN_LOOPBACK(ntohl(ip->ip_src.s_addr))))
			goto bad_pkt; /* XXX DROP!!! Bad packet. */

		/* IP check summ */
		if (filter->flags & UTP_FLT_FLAG_CSUM_IP_CHK) {
			if (0 == (m->m_pkthdr.csum_flags & CSUM_IP_CHECKED)) {
				if (0 == ((sizeof(struct ip) == l3_hlen) ?
				    in_cksum_hdr(ip) :
				    in_cksum_skip(m, (l2_hlen + l3_hlen), l2_hlen)))
					m->m_pkthdr.csum_flags |= CSUM_IP_VALID;
				m->m_pkthdr.csum_flags |= CSUM_IP_CHECKED;
			}
			if (0 == (m->m_pkthdr.csum_flags & CSUM_IP_VALID))
				goto bad_pkt; /* XXX DROP!!! Bad packet. */
		}

		/*
		 * Bypass:
		 * - fragments with offset (non first frag);
		 * - non UDP;
		 * - from/to broadcast (255.255.255.255);
		 * - from/to any (0.0.0.0);
		 * - from/to multicast;
		 * - small packets.
		 */
		if ((htons(IP_OFFMASK) & ip->ip_off) != 0 ||
		    IPPROTO_UDP != ip->ip_p ||
		    INADDR_BROADCAST == ip->ip_src.s_addr ||
		    INADDR_BROADCAST == ip->ip_dst.s_addr ||
		    INADDR_ANY == ip->ip_src.s_addr ||
		    INADDR_ANY == ip->ip_dst.s_addr ||
		    IN_MULTICAST(ntohl(ip->ip_dst.s_addr)) ||
		    IN_MULTICAST(ntohl(ip->ip_src.s_addr)) ||
		    (l3_hlen + sizeof(struct udphdr) + UTP_MIN_PKT_SIZE) > l3_len)
			goto bypass;
		/* In case of IP header with options, we haven't pulled up enough. */
		if (sizeof(struct ip) < l3_hlen) /* Options len. */
			pullup_len += (l3_hlen - sizeof(struct ip));

		/* LAYER 4 */
		/* Pool up udp header. */
		pullup_len += sizeof(struct udphdr);
		if (0 != (error = m_chk(&m, pullup_len)))
			goto mchk_err;
		udp = (struct udphdr *)(mtod(m, u_char *) + l2_hlen + l3_hlen);
		l4_len = ntohs(udp->uh_ulen);

		/* Basic packet checks, from udp_input.
		 * Drop (bypass) if:
		 * - Len less than size of udp_header;
		 * - Destination port of 0 is illegal, based on RFC768;
		 * - XXX "<" fragments -> "!=" no frags; Make mbuf data length 
		 *   reflect UDP length.
		 */
		if (sizeof(struct udphdr) > l4_len ||
		    0 == udp->uh_dport ||
		    (l3_len - l3_hlen) < l4_len)
			goto bad_pkt; /* XXX DROP!!! Bad packet. */

		/* UDP check summ */
		if (0 != udp->uh_sum && (filter->flags & UTP_FLT_FLAG_CSUM_UDP_CHK)) {
			uint16_t uh_sum;

			if (m->m_pkthdr.csum_flags & CSUM_DATA_VALID) {
				if (m->m_pkthdr.csum_flags & CSUM_PSEUDO_HDR)
					uh_sum = m->m_pkthdr.csum_data;
				else
					uh_sum = in_pseudo(ip->ip_src.s_addr,
					    ip->ip_dst.s_addr, htonl(l4_len +
					    m->m_pkthdr.csum_data + IPPROTO_UDP));
				uh_sum ^= 0xffff;
			} else {
				uh_sum = in_addword(
				    ~in_pseudo(ip->ip_src.s_addr, ip->ip_dst.s_addr,
					(udp->uh_ulen + htons(IPPROTO_UDP))),
				    in_cksum_skip(m, (l2_hlen + l3_len), 
				        (l2_hlen + l3_hlen)));
			}
			if (0 != uh_sum)
				goto bad_pkt; /* XXX DROP!!! Bad packet. */
		}

		l4_len = min((M_GET_WRITABLE_SIZE(m) - m->m_len), 
		    (ntohs(udp->uh_ulen) - sizeof(struct udphdr)));
		/* XXX need test with: kern.ipc.m_defragrandomfailures=1 */
		/*printf("%s: M_PULLUP_MAX = %u\n", __func__,
		    (uint32_t)M_PULLUP_MAX(m)); */
		if (NULL != m->m_next) {
			struct mbuf *m_new;
			m_new = m_defrag(m, M_NOWAIT);
			if (NULL == m_new) {
				//printf("%s: m_defrag fail\n", __func__);
				goto bypass;
			}
			m = m_new;//*/
			/* printf("%s: ip_len = %u, l3_hlen = %u, l4_len = %u, "
			    "pullup_len = %u\n", __func__, l3_len, l3_hlen, l4_len,
			    pullup_len);*/
			//m_print(m, 0);
			/*printf("%s: M_LEADINGSPACE = %u, M_TRAILINGSPACE = %u\n",
			    __func__, (uint32_t)M_LEADINGSPACE(m),
			    (uint32_t)M_TRAILINGSPACE(m));*/
		}
		//m_print(m, 0);

		/* Pool up udp data. */
		pullup_len += l4_len;
		//pullup_len = m->m_pkthdr.len;
		if (0 != (error = m_chk(&m, pullup_len))) {
			/*printf("%s: mchk_err: ip_len = %u, l3_hlen = %u, "
			    "l4_len = %u, pullup_len = %u\n", __func__, l3_len,
			    l3_hlen, l4_len, pullup_len);*/
			goto mchk_err;
		}
		ip = (struct ip *)(mtod(m, u_char *) + l2_hlen);
		daddr = (u_char *)&ip->ip_dst.s_addr;
		saddr = (u_char *)&ip->ip_src.s_addr;
		udp = (struct udphdr *)(mtod(m, u_char *) + l2_hlen + l3_hlen);
		udp_data = (u_char *)(udp + 1);
		break;
	case ETHERTYPE_IPV6:
		filter->stat.IPv6_Packets ++;
		goto bypass;
		break;
	default:
		goto bypass;
	}

	/* LAYER 5 */
	pkt_ver = uTPGetPacketVersion(udp_data, l4_len);
	if (-1 == pkt_ver)
		goto bypass; /* Not uTP packet. */
	connid = uTPGetConnID(pkt_ver, udp_data); /* Connection ID. */
	pkt_type = uTPGetFlagsType(pkt_ver, udp_data); /* Type (formerly flags). */
	seq_nr = uTPGetSeqNr(pkt_ver, udp_data); /* Sequence number. */
	ack_nr = uTPGetAckNr(pkt_ver, udp_data); /* Acknowledgment number. */

	/*
	 * Do not create "connection" for FIN and RST packets;
	 * Increase ConnID for SYN packets.
	 */
	switch (pkt_type) {
	case ST_SYN:	/* Connect ST_SYN. */
		/*
		 * SYN packet contain recv_connid - used for receive packets by
		 * initiator.
		 * Connection initiator use recv_connid+1 (recv_connid - from SYN)
		 * to send all other packets.
		 */
		connid ++;
		/* PASSTROUTH */
	case ST_DATA:	/* Data packet. */
	case ST_STATE:	/* State packet. Used to transmit an ACK with no data. */
		alloc_conn = 1;
		break;
	case ST_FIN:	/* Finalize the connection. This is the last packet. */
	case ST_RESET:	/* Terminate connection forcefully. */
	default:
		alloc_conn = 0;
	}

	/* LOCK HOST ITEM */
	hst_conn = hst_store_find_hst_conn(&priv->h_store, l3_p, saddr, daddr,
	    udp->uh_sport, udp->uh_dport, &hst_rev_state);
	if (NULL == hst_conn) { /* Create new session. */
		if (0 == alloc_conn)
			goto bypass;
		hst_rev_state = 0;
		hst_conn = hst_store_conn_alloc(&priv->h_store, l3_p, saddr, daddr,
		    udp->uh_sport, udp->uh_dport);
		switch (l3_p) {
		case ETHERTYPE_IP:
			priv->stat.ipv4_conns ++;
			break;
		case ETHERTYPE_IPV6:
			priv->stat.ipv6_conns ++;
			break;
		}
	}
	if (NULL == hst_conn) { /* Bypass uTP packet - no memory. */
		priv->stat.memoryFailures ++;
		goto bypass;
	}

	/* LOCKED */
	utp_conn = hst_store_find_utp_conn(hst_conn, hst_rev_state, pkt_ver, connid,
	    &utp_rev_state);
	if (NULL == utp_conn) { /* New connection, allocate. */
		if (0 == alloc_conn) {
			mtx_unlock(&hst_conn->rw_lock);
			goto bypass;
		}
		utp_rev_state = 0;
		utp_conn = hst_store_utp_conn_alloc(&priv->h_store, hst_conn,
		    hst_rev_state, pkt_ver, connid);
		/*printf("%s: Allocate  conn %u, pkt_type = %u, rev_state = %i, "
		    "pkt_ver = %i\n", __func__, connid, pkt_type, hst_rev_state,
		    pkt_ver); //*/
		/* Update stat. */
		switch (pkt_ver) {
		case 0:
			priv->stat.utpv0_conns ++;
			break;
		case 1:
			priv->stat.utpv1_conns ++;
			break;
		}
	}
	if (NULL == utp_conn) { /* Bypass uTP packet - no memory. */
		mtx_unlock(&hst_conn->rw_lock);
		priv->stat.memoryFailures ++;
		goto bypass;
	}
	time_now = time_second;
	hst_conn->last_update = time_now; /* Update time. */

	/* XXX analyze this before set. */
	//utp_conn->seq_nr = seq_nr; /* Sequence number. */
	//utp_conn->ack_nr = ack_nr; /* Acknowledgment number. */
	if (0 == utp_rev_state) {
		utp_conn->req_last = time_now; /* Update time. */
		utp_conn->req_count ++; /* Num of requests. */
		utp_conn->req_state |= CONN_FLAG_VAL(pkt_type); /* Remeber pkt type. */
	} else {
		utp_conn->repl_last = time_now; /* Update time. */
		utp_conn->repl_count ++; /* Num of replyes. */
		utp_conn->repl_state |= CONN_FLAG_VAL(pkt_type); /* Remeber pkt type. */
	}

	/* Syn send, rst send, fyn send, last req, last repl... */
	do_action = 0;
	switch (pkt_type) {
	case ST_DATA:	/* Data packet. */
	case ST_STATE:	/* State packet. Used to transmit an ACK with no data. */
		/* is ST_DATA or ST_SYN sended before from other side? */
		if (0 == utp_rev_state) {
			if (utp_conn->repl_state & CONN_F_CONNECTED &&
			    (time_now - utp_conn->repl_last) <
			    priv->cfg.conn_ack_timeout)
				do_action = 1;
		} else {
			if (utp_conn->req_state & CONN_F_CONNECTED &&
			    (time_now - utp_conn->req_last) <
			    priv->cfg.conn_ack_timeout)
				do_action = 1;
		}
		break;
	}
	//if ((utp_conn->req_count && utp_conn->repl_count) ||
	//    hst_conn->conns_count > 1)
	//	do_action = 1;

	mtx_unlock(&hst_conn->rw_lock);
	/* UNLOCK HOST ITEM */

	/* bypass uTP packet */
	if (0 == do_action)
		goto bypass;

	filter->stat.MatchedPackets ++;

	switch (filter->action) {
	case UTP_FLT_ACCTION_PASS:
		goto bypass;
		break;
	case UTP_FLT_ACCTION_DROP:
		goto drop;
		break;
	case UTP_FLT_ACCTION_PASS_MARK:
		/* LAYER 2 mark.*/
		/* NGM_UTP_ADD_FILTER not set this flag for L3 filter */
		if (filter->flags & UTP_FLT_FLAG_MARK_L2) {
			if (m->m_flags & M_VLANTAG) {
				m->m_pkthdr.ether_vtag = ((m->m_pkthdr.ether_vtag &
				    ~EVL_PRI_MASK) | filter->label_l2);
			} else if (sizeof(struct ether_vlan_header) == l2_hlen) {
				struct ether_vlan_header *evl;
				
				evl = mtod(m, struct ether_vlan_header *);
				if (htons(ETHERTYPE_VLAN) == evl->evl_encap_proto)
					evl->evl_tag = htons(((ntohs(evl->evl_tag) &
					    ~EVL_PRI_MASK) | filter->label_l2));
			}
		}
		/* LAYER 3 mark.*/
		if (filter->flags & UTP_FLT_FLAG_MARK_L3) {
			register uint8_t dscp_ecn;

			switch (l3_p) {
			case ETHERTYPE_IP:
				/*
				 * Allways recalc checksumm.
				 * "<< 8" = change byte at offset = 1
				 */
				dscp_ecn = ((ip->ip_tos & IPTOS_ECN_MASK) |
				    filter->label_l3);
				ip->ip_sum = csum_inc(ip->ip_sum, (ip->ip_tos << 8),
				    (dscp_ecn << 8));
				ip->ip_tos = dscp_ecn;
				break;
			case ETHERTYPE_IPV6:
				break;
			}
		}
		goto bypass;
		break;
	case UTP_FLT_ACCTION_SND_RST:
		break;
	}

	//goto drop; /* Drop uTP packet. */

	//printf("%s: sending RST pkt...\n", __func__);
	/* Replace packet to RST. */
	/* We must own the mbuf chain exclusively to modify it. */
	m = m_unshare(m, M_NOWAIT);// XXXXX
	if (NULL == m) {
		error = ENOMEM;
		printf("%s: m_unshare fail\n", __func__);
		goto mchk_err;
	}

	/* Reset csum flags and data. */
	m->m_pkthdr.csum_flags = 0;
	m->m_pkthdr.csum_data = 0;

	switch (l3_p) {
	case ETHERTYPE_IP:
		/* Remove ip options, if exist. */
		l4_len += (l3_hlen - sizeof(struct ip));
		l3_hlen = sizeof(struct ip);
		/* Update pointers. */
		ip = (struct ip *)(mtod(m, u_char *) + l2_hlen);
		udp = (struct udphdr *)(mtod(m, u_char *) + l2_hlen + l3_hlen);
		udp_data = (u_char *)(udp + 1);
		
		/* Generate uTP RST (new UDP data). */
		l4_len = uTPGenRSTPacket(pkt_ver, connid, ack_nr, seq_nr, udp_data,
		    l4_len);

		/* Update UDP and IP headers. */
		l4_len += sizeof(struct udphdr);
		l3_len = (l3_hlen + l4_len);
		
		/* Update UDP hdr. */
		udp->uh_ulen = htons(l4_len);
		udp->uh_sum = 0;
		if (0 == (filter->flags & UTP_FLT_FLAG_CSUM_UDP_NO_CSUMM)) {
			/*
			 * Equial to:
			 * udp->uh_sum = in_pseudo(...);
			 * udp->uh_sum = in_cksum_skip(...); - cover "uh_sum" from 
			 * previuos line.
			 */
			udp->uh_sum = in_addword(
			    ~in_pseudo(ip->ip_src.s_addr, ip->ip_dst.s_addr,
				(udp->uh_ulen + htons(IPPROTO_UDP))),
			    in_cksum_skip(m, (l2_hlen + l3_len),
			        (l2_hlen + l3_hlen)));
			//if (0 == udp->uh_sum) /* XXX check this */
			//	udp->uh_sum = 0xffff;
			m->m_pkthdr.csum_flags |= (CSUM_PSEUDO_HDR | CSUM_DATA_VALID);
			m->m_pkthdr.csum_data = 0xffff;
		}

		/* Update IP hdr. */
		//ip->ip_v = IPVERSION; /* No change. */
		ip->ip_hl = (sizeof(struct ip) >> 2); /* Overwrite ip hdr len. */
		//ip->ip_tos = 0; /* No change. */
		ip->ip_len = htons(l3_len);
		//ip->ip_id = ; /* No change. */
		ip->ip_off = 0; /* set to not fragmented */
		//ip->ip_ttl = ; /* No change. */
		//ip->ip_p = IPPROTO_UDP; /* No change. */
		ip->ip_sum = 0;
		ip->ip_sum = (sizeof(struct ip) == l3_hlen) ?
		    in_cksum_hdr(ip) :
		    in_cksum_skip(m, (l2_hlen + l3_hlen), l2_hlen);
		m->m_pkthdr.csum_flags |= (CSUM_IP_CHECKED | CSUM_IP_VALID);

		/* Truncate packet length if required. */
		m_adj(m, -(m->m_pkthdr.len - (l2_hlen + l3_len)));

		goto fwd_pkt;
		break;
	case ETHERTYPE_IPV6:
		goto bypass;
		break;
	}


bad_pkt: /* Default bypass bad packets. */
	filter->stat.recvErrors ++;

bypass:	/* Passtrouth packet. */
	if (filter->flags & UTP_FLT_FLAG_NO_PASS)
		goto drop;
	filter->stat.BypassPackets ++;

fwd_pkt:
	NG_FWD_NEW_DATA(error, item, hpriv->h_pass, m);
	if (0 != error)
		filter->stat.xmitErrors ++;
	return (error);

drop:
	filter->stat.DroppedPackets ++;
	m_freem(m);
	NG_FREE_ITEM(item);
	return (error);

mchk_err:
	if (EINVAL == error)
		goto bypass; /* Bypass packet. */
	priv->stat.memoryFailures ++;
	NG_FREE_ITEM(item);
	return (error);
}

